
import { _decorator, ScrollView, UITransform, v3, Node, Rect, rect, Vec3 } from 'cc';
import ScroViewItem from './ScroViewItem';

const { ccclass, property } = _decorator;

/**
 * 用法：
 *
 *      1. 将本组件挂载在节点上即可，和正常ScrollView使用一致
 *
 * 原理：
 *
 *      1. 滚动时，判断子节点是否进入了/离开了可视区域
 *      2. 根据结果回调对应事件，在可以实现类似以下功能：
 *          * 控制可视区域Item显示（透明度改为 255 ），非可视区域Item隐藏（透明度改为 0 ）
 */
@ccclass
export default class ScroView extends ScrollView {

    @property({ tooltip: "是否计算在可视区域中Item的相对位置（可能会相对耗性能）" })
    caculatePosition: boolean = false;
    callfunc: Function;

    onEnable() {
        super.onEnable();

        this.node.on("scrolling", this._onScrollingDrawCallOpt, this);
        this.node.on("ended", this._onScrollingDrawCallOpt, this);

        this.optDc();
    }

    onDisable() {
        this.node.off("ended", this._onScrollingDrawCallOpt, this);
        this.node.off("scrolling", this._onScrollingDrawCallOpt, this);

        super.onDisable();
    }

    /**
     * 
     * @param numb 分帧加载的数量
     * @param callfunc 分帧加载完成回调
     */
    async onInitView(numb: number, callfunc: Function, isRemove: boolean = false) {
        this.callfunc = callfunc;
        if (isRemove)
            this.content.destroyAllChildren();
        await this.executePreFrame(this._getItemGenerator(numb), 1);
        // 在创建好子节点之后，先手动调用一次DC优化，触发当前在可视区域内的节点的进入逻辑
        // 后续的ScrollView滚动时，内部自动回调
        this.optDc();
    }

    /**
     * 分帧执行 Generator 逻辑
     * @param generator 生成器
     * @param duration 持续时间（ms），每次执行 Generator 的操作时，最长可持续执行时长。假设值为8ms，那么表示1帧（总共16ms）下，分出8ms时间给此逻辑执行
     */
    private executePreFrame(generator: Generator, duration: number) {
        return new Promise((resolve, reject) => {
            let gen = generator;
            // 创建执行函数
            let execute = () => {
                // 执行之前，先记录开始时间
                let startTime = new Date().getTime();
                // 然后一直从 Generator 中获取已经拆分好的代码段出来执行
                for (let iter = gen.next(); ; iter = gen.next()) {
                    // 判断是否已经执行完所有 Generator 的小代码段，如果是的话，那么就表示任务完成
                    if (iter == null || iter.done) {
                        resolve(0);
                        return;
                    }

                    // 每执行完一段小代码段，都检查一下是否已经超过我们分配的本帧，这些小代码端的最大可执行时间
                    if (new Date().getTime() - startTime > duration) {
                        // 如果超过了，那么本帧就不在执行，开定时器，让下一帧再执行
                        this.scheduleOnce(() => { execute(); });
                        return;
                    }
                }
            };
            // 运行执行函数
            execute();
        });
    }

    /**
     * 执行次数
     * @param length 需要渲染的item
     */
    private *_getItemGenerator(length: number) {
        if (!this.callfunc) return console.log("渲染节点出问题!");
        for (let i = 0; i < length; i++) {
            yield this.callfunc(i);
        }
    }

    private _onScrollingDrawCallOpt() {
        if (!this.content.children || this.content.children.length == 0) {
            return;
        }
        this.optDc();
    }

    public optDc() {
        ScroView.optDc(this, this.caculatePosition);
    }

    /**
     * 优化 ScrollView Content 节点 DC，可以手动调用
     *
     * 具体为
     *
     * 1. 进入ScrollView可视区域是，回调对应 Content 子节点上挂载的 ScollViewPlusItem 组件的 onEnterScorllViewEvents 数组事件
     * 2. 退出ScrollView可视区域是，回调对应 Content 子节点上挂载的 ScollViewPlusItem 组件的 onExitScorllViewEvents 数组事件
     */
    private static optDc(scrollView: ScrollView, caculatePosition: boolean): void {
        // 获取 ScrollView Node 的左下角坐标在世界坐标系中的坐标
        let nTransform = scrollView.node.getComponent(UITransform);

        let uiTransform = scrollView.node.parent.getComponent(UITransform);
        let svLeftBottomPoint: Vec3 = uiTransform.convertToWorldSpaceAR(
            v3(
                scrollView.node.position.x - nTransform.anchorX * nTransform.width,
                scrollView.node.position.y - nTransform.anchorY * nTransform.height
            )
        );

        // 求出 ScrollView 可视区域在世界坐标系中的矩形（碰撞盒）
        let svBBoxRect: Rect = rect(svLeftBottomPoint.x, svLeftBottomPoint.y, nTransform.width, nTransform.height);
 
        // 遍历 ScrollView Content 内容节点的子节点，对每个子节点的包围盒做和 ScrollView 可视区域包围盒做碰撞判断
        scrollView.content.children.forEach((childNode: Node) => {
            // 没有绑定指定组件的子节点不处理
            let itemComponent = childNode.getComponent(ScroViewItem);
            if (itemComponent == null) {
                return;
            }

            // 如果相交了，那么就显示，否则就隐藏
            let cTransform = childNode.getComponent(UITransform);
            let childNodeBBox = cTransform.getBoundingBoxToWorld();
            if (childNodeBBox.intersects(svBBoxRect)) {
                itemComponent.publishOnEnterScrollView();
                if (caculatePosition) {
                    itemComponent.publishOnPositionChange(
                        (childNodeBBox.x - (svBBoxRect.x - childNodeBBox.width / 2)) / (childNodeBBox.width + svBBoxRect.width),
                        (childNodeBBox.y - (svBBoxRect.y - childNodeBBox.height / 2)) / (childNodeBBox.height + svBBoxRect.height)
                    );
                }
            }
            else {
                itemComponent.publishOnExitScrollView();
            }
        });
    }
}
